Imports Microsoft.VisualBasic
Imports System
Imports System.IO

Namespace NeuralClassification
    Friend Class NeuralClassificationProgram
        Private Shared rnd As Random = Nothing

        Shared Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbCrLf & "Begin neural network classification demo" & vbCrLf)
                Console.WriteLine("Goal is to predict/classify color based on four numeric inputs" & vbCrLf)
                rnd = New Random(159) ' 159 makes 'good' output

                Console.WriteLine("Creating 100 lines of raw data")
                Dim dataFile As String = "..\..\colors.txt"
                MakeData(dataFile, 100)

                Console.WriteLine(vbCrLf & "First few rows of raw data file are:")
                Helpers.ShowTextFile(dataFile, 4)

                Dim trainMatrix()() As Double = Nothing
                Dim testMatrix()() As Double = Nothing
                Console.WriteLine(vbCrLf & "Generating train and test matrices using an 80%-20% split")
                MakeTrainAndTest(dataFile, trainMatrix, testMatrix)

                Console.WriteLine(vbCrLf & "First few rows of training matrix are:")
                Helpers.ShowMatrix(trainMatrix, 5)

                Console.WriteLine(vbCrLf & "Creating 4-input 5-hidden 3-output neural network")
                Dim nn As New NeuralNetwork(4, 5, 3)

                Console.WriteLine("Training to find best neural network weights using PSO with cross entropy error")
                Dim bestWeights() As Double = nn.Train(trainMatrix)
                Console.WriteLine(vbCrLf & "Best weights found:")
                Helpers.ShowVector(bestWeights, 2, True)

                Console.WriteLine(vbCrLf & "Loading best weights into neural network")
                nn.SetWeights(bestWeights)

                Console.WriteLine(vbCrLf & "Analyzing the neural network accuracy on the test data" & vbCrLf)
                Dim accuracy As Double = nn.Test(testMatrix)
                Console.WriteLine("Prediction accuracy = " & accuracy.ToString("F4"))

                Console.WriteLine(vbCrLf & "End neural network classification demo" & vbCrLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main()

        ' --------------------------------------------------------------------------------------------

        Private Shared Sub MakeData(ByVal dataFile As String, ByVal numLines As Integer)
            Dim weights() As Double = {-0.1, 0.2, -0.3, 0.4, -0.5, 0.6, -0.7, 0.8, -0.9, 1.0, -1.1, 1.2, -1.3, 1.4, -1.5, 1.6, -1.7, 1.8, -1.9, 2.0, -0.5, 0.6, -0.7, 0.8, -0.9, 1.5, -1.4, 1.3, -1.2, 1.1, -1.0, 0.9, -0.8, 0.7, -0.6, 0.5, -0.4, 0.3, -0.2, 0.1, 0.1, -0.3, 0.6}

            Dim nn As New NeuralNetwork(4, 5, 3)
            nn.SetWeights(weights)

            Dim ofs As New FileStream(dataFile, FileMode.Create)
            Dim sw As New StreamWriter(ofs)

            For i  = 0 To numLines - 1
                Dim inputs(3) As Double
                For j  = 0 To inputs.Length - 1
                    inputs(j) = rnd.Next(1, 10)
                Next j

                Dim outputs() As Double = nn.ComputeOutputs(inputs)

                Dim color As String = ""
                Dim idx As Integer = Helpers.IndexOfLargest(outputs)
                If idx = 0 Then
                    color = "red"
                ElseIf idx = 1 Then
                    color = "green"
                ElseIf idx = 2 Then
                    color = "blue"
                End If

                sw.WriteLine(inputs(0).ToString("F1") & " " & inputs(1).ToString("F1") & " " & inputs(2).ToString("F1") & " " & inputs(3).ToString("F1") & " " & color)
            Next i
            sw.Close()
            ofs.Close()
        End Sub ' MakeData

        Private Shared Sub MakeTrainAndTest(ByVal file As String, <System.Runtime.InteropServices.Out> ByRef trainMatrix()() As Double, <System.Runtime.InteropServices.Out> ByRef testMatrix()() As Double)
            Dim numLines As Integer = 0
            Dim ifs As New FileStream(file, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Do While sr.ReadLine IsNot Nothing
                numLines += 1
            Loop
            sr.Close()
            ifs.Close()

            Dim numTrain As Integer = CInt(Fix(0.8 * numLines))
            Dim numTest As Integer = numLines - numTrain

            Dim allData(numLines - 1)() As Double ' could use Helpers.MakeMatrix here
            For i  = 0 To allData.Length - 1
                allData(i) = New Double(6) {} ' (x0, x1, x2, x3), (y0, y1, y2)
            Next i

            Dim line As String = ""
            Dim tokens() As String = Nothing
            ifs = New FileStream(file, FileMode.Open)
            sr = New StreamReader(ifs)
            Dim row As Integer = 0
            line = sr.ReadLine
            Do While line IsNot Nothing
                tokens = line.Split(" "c)
                allData(row)(0) = Double.Parse(tokens(0))
                allData(row)(1) = Double.Parse(tokens(1))
                allData(row)(2) = Double.Parse(tokens(2))
                allData(row)(3) = Double.Parse(tokens(3))

                For i  = 0 To 3
                    allData(row)(i) = 0.25 * allData(row)(i) - 1.25 ' scale input data to -1.0, +1.0
                Next i

                If tokens(4) = "red" Then
                    allData(row)(4) = 1.0
                    allData(row)(5) = 0.0
                    allData(row)(6) = 0.0
                ElseIf tokens(4) = "green" Then
                    allData(row)(4) = 0.0
                    allData(row)(5) = 1.0
                    allData(row)(6) = 0.0
                ElseIf tokens(4) = "blue" Then
                    allData(row)(4) = 0.0
                    allData(row)(5) = 0.0
                    allData(row)(6) = 1.0
                End If
                row += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()

            Helpers.ShuffleRows(allData)

            trainMatrix = Helpers.MakeMatrix(numTrain, 7)
            testMatrix = Helpers.MakeMatrix(numTest, 7)

            For i  = 0 To numTrain - 1
                allData(i).CopyTo(trainMatrix(i), 0)
            Next i

            For i  = 0 To numTest - 1
                allData(i + numTrain).CopyTo(testMatrix(i), 0)
            Next i
        End Sub ' MakeTrainAndTest

        ' --------------------------------------------------------------------------------------------
    End Class ' class NeuralClassificationProgram

    ' ==============================================================================================

    Friend Class NeuralNetwork
        Private numInput As Integer
        Private numHidden As Integer
        Private numOutput As Integer

        Private inputs() As Double
        Private ihWeights()() As Double ' input-to-hidden
        Private ihSums() As Double
        Private ihBiases() As Double
        Private ihOutputs() As Double
        Private hoWeights()() As Double ' hidden-to-output
        Private hoSums() As Double
        Private hoBiases() As Double
        Private outputs() As Double

        Private Shared rnd As Random = Nothing

        Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer)
            Me.numInput = numInput
            Me.numHidden = numHidden
            Me.numOutput = numOutput

            inputs = New Double(numInput - 1) {}
            ihWeights = Helpers.MakeMatrix(numInput, numHidden)
            ihSums = New Double(numHidden - 1) {}
            ihBiases = New Double(numHidden - 1) {}
            ihOutputs = New Double(numHidden - 1) {}
            hoWeights = Helpers.MakeMatrix(numHidden, numOutput)
            hoSums = New Double(numOutput - 1) {}
            hoBiases = New Double(numOutput - 1) {}
            outputs = New Double(numOutput - 1) {}

            rnd = New Random(0)
        End Sub

        Public Sub SetWeights(ByVal weights() As Double)
            Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
            If weights.Length <> numWeights Then
                Throw New Exception("The weights array length: " & weights.Length & " does not match the total number of weights and biases: " & numWeights)
            End If

            Dim k As Integer = 0 ' points into weights param

            For i  = 0 To numInput - 1
                For j  = 0 To numHidden - 1
                    ihWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i

            For i  = 0 To numHidden - 1
                ihBiases(i) = weights(k)
                k += 1
            Next i

            For i  = 0 To numHidden - 1
                For j  = 0 To numOutput - 1
                    hoWeights(i)(j) = weights(k)
                    k += 1
                Next j
            Next i

            For i  = 0 To numOutput - 1
                hoBiases(i) = weights(k)
                k += 1
            Next i
        End Sub

        Public Function ComputeOutputs(ByVal currInputs() As Double) As Double()
            If inputs.Length <> numInput Then
                Throw New Exception("Inputs array length " & inputs.Length & " does not match NN numInput value " & numInput)
            End If

            For i  = 0 To numHidden - 1
                Me.ihSums(i) = 0.0
            Next i
            'For i  = 0 To numHidden - 1
            '  Me.ihOutputsi = 0.0
            For i  = 0 To numOutput - 1
                Me.hoSums(i) = 0.0
            Next i
            'For i = 0 To numOutput - 1
            '  Me.outputsi = 0.0

            For i  = 0 To currInputs.Length - 1 ' copy
                Me.inputs(i) = currInputs(i)
            Next i

            'Console.WriteLine("Inputs:")
            'ShowVector(Me.inputs)

            'Console.WriteLine("input-to-hidden weights:")
            'ShowMatrix(Me.ihWeights)

            For j  = 0 To numHidden - 1 ' compute input-to-hidden sums
                For i  = 0 To numInput - 1
                    ihSums(j) += Me.inputs(i) * ihWeights(i)(j)
                Next i
            Next j

            'Console.WriteLine("input-to-hidden sums:")
            'ShowVector(Me.ihSums)

            'Console.WriteLine("input-to-hidden biases:")
            'ShowVector(ihBiases)

            For i  = 0 To numHidden - 1 ' add biases to input-to-hidden sums
                ihSums(i) += ihBiases(i)
            Next i

            'Console.WriteLine("input-to-hidden sums after adding biases:")
            'ShowVector(Me.ihSums)

            For i  = 0 To numHidden - 1 ' determine input-to-hidden output
                'ihOutputsi = StepFunction(ihSumsi); // step function
                ihOutputs(i) = SigmoidFunction(ihSums(i))
            Next i
            'ihOutputsi = TanhFunction(ihSumsi)

            'Console.WriteLine("input-to-hidden outputs after sigmoid:")
            'ShowVector(Me.ihOutputs)

            'Console.WriteLine("hidden-to-output weights:")
            'ShowMatrix(hoWeights)

            For j  = 0 To numOutput - 1 ' compute hidden-to-output sums
                For i  = 0 To numHidden - 1
                    hoSums(j) += ihOutputs(i) * hoWeights(i)(j)
                Next i
            Next j

            'Console.WriteLine("hidden-to-output sums:")
            'ShowVector(hoSums)

            'Console.WriteLine("hidden-to-output biases:")
            'ShowVector(Me.hoBiases)

            For i  = 0 To numOutput - 1 ' add biases to input-to-hidden sums
                hoSums(i) += hoBiases(i)
            Next i

            'Console.WriteLine("hidden-to-output sums after adding biases:")
            'ShowVector(Me.hoSums)

            'For i = 0 To numOutput - 1   // determine hidden-to-output result
            '  Me.outputsi = SigmoidFunction(hoSumsi);  // step function
            'Next i 
            'Dim result(numOutput - 1) As Double
            'Me.outputs.CopyTo(result, 0)
            'return result

            Dim result() As Double = Softmax(hoSums)

            result.CopyTo(Me.outputs, 0)

            'Console.WriteLine("outputs after softmaxing:")
            'ShowVector(result)

            'Console.ReadLine

            'Dim result() As Double = Hardmax(hoSums)
            Return result
        End Function ' ComputeOutputs

        'Private Shared Function StepFunction(ByVal x As Double) As Double
        '    If x > 0.0 Then
        '        Return 1.0
        '    Else
        '        Return 0.0
        'End Function
        Private Shared Function SigmoidFunction(ByVal x As Double) As Double
            If x < -45.0 Then
                Return 0.0
            ElseIf x > 45.0 Then
                Return 1.0
            Else
                Return 1.0 / (1.0 + Math.Exp(-x))
            End If
        End Function

        Private Shared Function Softmax(ByVal hoSums() As Double) As Double()
            ' determine max
            Dim max As Double = hoSums(0)
            For i = 0 To hoSums.Length - 1
                If hoSums(i) > max Then
                    max = hoSums(i)
                End If
            Next i

            ' determine scaling factor (sum of exp(eachval - max)
            Dim scale As Double = 0.0
            For i = 0 To hoSums.Length - 1
                scale += Math.Exp(hoSums(i) - max)
            Next i

            Dim result(hoSums.Length - 1) As Double
            For i = 0 To hoSums.Length - 1
                result(i) = Math.Exp(hoSums(i) - max) / scale
            Next i

            Return result
        End Function

        Public Function Train(ByVal trainMatrix()() As Double) As Double() ' seek and return the best weights
            Dim numWeights As Integer = (Me.numInput * Me.numHidden) + (Me.numHidden * Me.numOutput) + Me.numHidden + Me.numOutput
            'Dim currWeights(numWeights - 1) As Double

            ' use PSO to seek best weights
            Dim numberParticles As Integer = 10
            Dim numberIterations As Integer = 500
            Dim iteration As Integer = 0
            Dim [Dim] As Integer = numWeights ' number of values to solve for
            Dim minX As Double = -5.0 ' for each weight
            Dim maxX As Double = 5.0

            Dim swarm(numberParticles - 1) As Particle
            Dim bestGlobalPosition([Dim] - 1) As Double ' best solution found by any particle in the swarm. implicit initialization to all 0.0
            Dim bestGlobalFitness As Double = Double.MaxValue ' smaller values better

            Dim minV As Double = -0.1 * maxX ' velocities
            Dim maxV As Double = 0.1 * maxX

            For i = 0 To swarm.Length - 1 ' initialize each Particle in the swarm with random positions and velocities
                Dim randomPosition([Dim] - 1) As Double
                For j = 0 To randomPosition.Length - 1
                    Dim lo As Double = minX
                    Dim hi As Double = maxX
                    randomPosition(j) = (hi - lo) * rnd.NextDouble + lo
                Next j

                Dim fitness As Double = CrossEntropy(trainMatrix, randomPosition) ' smaller values better
                Dim randomVelocity([Dim] - 1) As Double

                For j = 0 To randomVelocity.Length - 1
                    Dim lo As Double = -1.0 * Math.Abs(maxX - minX)
                    Dim hi As Double = Math.Abs(maxX - minX)
                    randomVelocity(j) = (hi - lo) * rnd.NextDouble + lo
                Next j
                swarm(i) = New Particle(randomPosition, fitness, randomVelocity, randomPosition, fitness)

                ' does current Particle have global best position/solution?
                If swarm(i).fitness < bestGlobalFitness Then
                    bestGlobalFitness = swarm(i).fitness
                    swarm(i).position.CopyTo(bestGlobalPosition, 0)
                End If
            Next i ' initialization

            Dim w As Double = 0.729 ' inertia weight.
            Dim c1 As Double = 1.49445 ' cognitive/local weight
            Dim c2 As Double = 1.49445 ' social/global weight
            Dim r1, r2 As Double ' cognitive and social randomizations

            Console.WriteLine("Entering main PSO weight estimation processing loop")
            Do While iteration < numberIterations
                iteration += 1
                Dim newVelocity([Dim] - 1) As Double
                Dim newPosition([Dim] - 1) As Double
                Dim newFitness As Double

                For i = 0 To swarm.Length - 1 ' each Particle
                    Dim currP As Particle = swarm(i)

                    For j = 0 To currP.velocity.Length - 1 ' each x value of the velocity
                        r1 = rnd.NextDouble
                        r2 = rnd.NextDouble

                        newVelocity(j) = (w * currP.velocity(j)) + (c1 * r1 * (currP.bestPosition(j) - currP.position(j))) + (c2 * r2 * (bestGlobalPosition(j) - currP.position(j))) ' new velocity depends on old velocity, best position of parrticle, and best position of any particle

                        If newVelocity(j) < minV Then
                            newVelocity(j) = minV
                        ElseIf newVelocity(j) > maxV Then
                            newVelocity(j) = maxV ' crude way to keep velocity in range
                        End If
                    Next j

                    newVelocity.CopyTo(currP.velocity, 0)

                    For j = 0 To currP.position.Length - 1
                        newPosition(j) = currP.position(j) + newVelocity(j) ' compute new position
                        If newPosition(j) < minX Then
                            newPosition(j) = minX
                        ElseIf newPosition(j) > maxX Then
                            newPosition(j) = maxX
                        End If
                    Next j

                    newPosition.CopyTo(currP.position, 0)

                    newFitness = CrossEntropy(trainMatrix, newPosition) ' compute error of the new position
                    currP.fitness = newFitness

                    If newFitness < currP.bestFitness Then ' new particle best?
                        newPosition.CopyTo(currP.bestPosition, 0)
                        currP.bestFitness = newFitness
                    End If

                    If newFitness < bestGlobalFitness Then ' new global best?
                        newPosition.CopyTo(bestGlobalPosition, 0)
                        bestGlobalFitness = newFitness
                    End If
                Next i ' each Particle

                'Console.WriteLine(swarm0.ToString)
                'Console.ReadLine
            Loop ' while

            Console.WriteLine("Processing complete")
            Console.Write("Final best (smallest) cross entropy error = ")
            Console.WriteLine(bestGlobalFitness.ToString("F4"))

            Return bestGlobalPosition
        End Function ' Train

        Private Function CrossEntropy(ByVal trainData()() As Double, ByVal weights() As Double) As Double ' (sum) Cross Entropy
            ' how good (cross entropy) are weights? CrossEntropy is error so smaller values are better
            Me.SetWeights(weights) ' load the weights and biases to examine

            Dim sce As Double = 0.0 ' sum of cross entropy

            For i = 0 To trainData.Length - 1 ' walk thru each training case. looks like (6.9 3.2 5.7 2.3) (0 0 1) where the parens are not really there
                Dim currInputs(3) As Double
                currInputs(0) = trainData(i)(0)
                currInputs(1) = trainData(i)(1)
                currInputs(2) = trainData(i)(2)
                currInputs(3) = trainData(i)(3)
                Dim currExpected(2) As Double ' not really necessary
                currExpected(0) = trainData(i)(4)
                currExpected(1) = trainData(i)(5)
                currExpected(2) = trainData(i)(6)

                Dim currOutputs() As Double = Me.ComputeOutputs(currInputs) ' run the jnputs through the neural network

                ' compute ln of each nn output (and the sum)
                Dim currSum As Double = 0.0
                For j = 0 To currOutputs.Length - 1
                    If currExpected(j) <> 0.0 Then
                        currSum += currExpected(j) * Math.Log(currOutputs(j))
                    End If
                Next j
                sce += currSum ' accumulate
            Next i
            Return -sce
        End Function ' CrossEntropy

        Public Function Test(ByVal testMatrix()() As Double) As Double ' returns the accuracy (percent correct predictions)
            ' assumes that weights have been set using SetWeights
            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0

            For i = 0 To testMatrix.Length - 1 ' walk thru each test case. looks like (6.9 3.2 5.7 2.3) (0 0 1) where the parens are not really there
                Dim currInputs(3) As Double
                currInputs(0) = testMatrix(i)(0)
                currInputs(1) = testMatrix(i)(1)
                currInputs(2) = testMatrix(i)(2)
                currInputs(3) = testMatrix(i)(3)
                Dim currOutputs(2) As Double ' not really necessary
                currOutputs(0) = testMatrix(i)(4)
                currOutputs(1) = testMatrix(i)(5)
                currOutputs(2) = testMatrix(i)(6)
                Dim currPredicted() As Double = Me.ComputeOutputs(currInputs) ' outputs are in softmax form -- each between 0.0, 1.0 representing a prob and summing to 1.0

                'ShowVector(currInputs)
                'ShowVector(currOutputs)
                'ShowVector(currPredicted)

                ' use winner-takes all -- highest prob of the prediction
                Dim indexOfLargest As Integer = Helpers.IndexOfLargest(currPredicted)

                If i <= 3 Then ' just a few for demo purposes
                    Console.WriteLine("-----------------------------------")
                    Console.Write("Input:     ")
                    Helpers.ShowVector(currInputs, 2, True)
                    Console.Write("Output:    ")
                    Helpers.ShowVector(currOutputs, 1, False)
                    If currOutputs(0) = 1.0 Then
                        Console.WriteLine(" (red)")
                    ElseIf currOutputs(1) = 1.0 Then
                        Console.WriteLine(" (green)")
                    Else
                        Console.WriteLine(" (blue)")
                    End If
                    Console.Write("Predicted: ")
                    Helpers.ShowVector(currPredicted, 1, False)
                    If indexOfLargest = 0 Then
                        Console.WriteLine(" (red)")
                    ElseIf indexOfLargest = 1 Then
                        Console.WriteLine(" (green)")
                    Else
                        Console.WriteLine(" (blue)")
                    End If

                    If currOutputs(indexOfLargest) = 1 Then
                        Console.WriteLine("correct")
                    Else
                        Console.WriteLine("wrong")
                    End If
                    Console.WriteLine("-----------------------------------")
                End If

                If currOutputs(indexOfLargest) = 1 Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If

                'Console.ReadLine
            Next i
            Console.WriteLine(". . .")

            Dim percentCorrect As Double = (numCorrect * 1.0) / (numCorrect + numWrong)
            Console.WriteLine(vbCrLf & "Correct = " & numCorrect)
            Console.WriteLine("Wrong = " & numWrong)

            Return percentCorrect
        End Function ' Test
    End Class ' class NeuralNetwork

    ' ==============================================================================================

    Public Class Helpers
        Private Shared rnd As New Random(0)

        Public Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows - 1)() As Double
            For i  = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Public Shared Sub ShuffleRows(ByVal matrix()() As Double)
            For i  = 0 To matrix.Length - 1
                Dim r As Integer = rnd.Next(i, matrix.Length)
                Dim tmp() As Double = matrix(r)
                matrix(r) = matrix(i)
                matrix(i) = tmp
            Next i
        End Sub

        Public Shared Function IndexOfLargest(ByVal vector() As Double) As Integer
            Dim _indexOfLargest As Integer = 0
            Dim maxVal As Double = vector(0)
            For i  = 0 To vector.Length - 1
                If vector(i) > maxVal Then
                    maxVal = vector(i)
                    _indexOfLargest = i
                End If
            Next i
            Return _indexOfLargest
        End Function

        Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal newLine As Boolean)
            Dim fmt As String = "F" & decimals
            For i  = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod 12 = 0 Then
                    Console.WriteLine("")
                End If
                If vector(i) >= 0.0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString(fmt) & " ")
            Next i
            If newLine = True Then
                Console.WriteLine("")
            End If
        End Sub

        Public Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer)
            Dim ct As Integer = 0
            If numRows = -1 Then
                numRows = Integer.MaxValue
            End If
            Dim i As Integer = 0
            Do While i < matrix.Length AndAlso ct < numRows
                For j  = 0 To matrix(0).Length - 1
                    If matrix(i)(j) >= 0.0 Then
                        Console.Write(" ")
                    End If
                    If j = 4 Then
                        Console.Write("-> ")
                    End If
                    Console.Write(matrix(i)(j).ToString("F2") & " ")
                Next j
                Console.WriteLine("")
                ct += 1
                i += 1
            Loop
            Console.WriteLine("")
        End Sub

        Public Shared Sub ShowTextFile(ByVal textFile As String, ByVal numLines As Integer)
            Dim ifs As New FileStream(textFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim ct As Integer = 0
            line = sr.ReadLine
            Do While line IsNot Nothing AndAlso ct < numLines
                Console.WriteLine(line)
                ct += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()
        End Sub
    End Class ' class Helpers

    ' ==============================================================================================

    Public Class Particle
        Public position() As Double ' equivalent to x-Values and/or solution
        Public fitness As Double
        Public velocity() As Double

        Public bestPosition() As Double ' best position found so far by this Particle
        Public bestFitness As Double

        Public Sub New(ByVal position() As Double, ByVal fitness As Double, ByVal velocity() As Double, ByVal bestPosition() As Double, ByVal bestFitness As Double)
            Me.position = New Double(position.Length - 1) {}
            position.CopyTo(Me.position, 0)
            Me.fitness = fitness
            Me.velocity = New Double(velocity.Length - 1) {}
            velocity.CopyTo(Me.velocity, 0)
            Me.bestPosition = New Double(bestPosition.Length - 1) {}
            bestPosition.CopyTo(Me.bestPosition, 0)
            Me.bestFitness = bestFitness
        End Sub

        Public Overrides Function ToString() As String
            Dim s As String = ""
            s &= "==========================" & vbCrLf
            s &= "Position: "
            For i  = 0 To Me.position.Length - 1
                s &= Me.position(i).ToString("F2") & " "
            Next i
            s &= vbCrLf
            s &= "Fitness = " & Me.fitness.ToString("F4") & vbCrLf
            s &= "Velocity: "
            For i  = 0 To Me.velocity.Length - 1
                s &= Me.velocity(i).ToString("F2") & " "
            Next i
            s &= vbCrLf
            s &= "Best Position: "
            For i  = 0 To Me.bestPosition.Length - 1
                s &= Me.bestPosition(i).ToString("F2") & " "
            Next i
            s &= vbCrLf
            s &= "Best Fitness = " & Me.bestFitness.ToString("F4") & vbCrLf
            s &= "==========================" & vbCrLf
            Return s
        End Function
    End Class ' class Particle
End Namespace ' ns
